@@ -40,9 +40,6 @@
 #include "win32/param.h"
 #include "win32/winutil.h"
 #define GET_DL_ERROR()	php_win_err()
-#elif defined(NETWARE)
-#include <sys/param.h>
-#define GET_DL_ERROR()	dlerror()
 #else
 #include <sys/param.h>
 #define GET_DL_ERROR()	DL_ERROR()
@@ -56,9 +53,9 @@ PHPAPI PHP_FUNCTION(dl)
 	char *filename;
 	size_t filename_len;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "s", &filename, &filename_len) == FAILURE) {
-		return;
-	}
+	ZEND_PARSE_PARAMETERS_START(1, 1)
+		Z_PARAM_STRING(filename, filename_len)
+	ZEND_PARSE_PARAMETERS_END();
 
 	if (!PG(enable_dl)) {
 		php_error_docref(NULL, E_WARNING, "Dynamically loaded extensions aren't enabled");
@@ -79,11 +76,31 @@ PHPAPI PHP_FUNCTION(dl)
 
 #if defined(HAVE_LIBDL)
 
-#ifdef ZTS
-#define USING_ZTS 1
+/* {{{ php_load_shlib
+ */
+PHPAPI void *php_load_shlib(char *path, char **errp)
+{
+	void *handle;
+	char *err;
+
+	handle = DL_LOAD(path);
+	if (!handle) {
+		err = GET_DL_ERROR();
+#ifdef PHP_WIN32
+		if (err && (*err)) {
+			(*errp)=estrdup(err);
+			LocalFree(err);
+		} else {
+			(*errp) = estrdup("<No message>");
+		}
 #else
-#define USING_ZTS 0
+		(*errp) = estrdup(err);
+		GET_DL_ERROR(); /* free the buffer storing the error */
 #endif
+	}
+	return handle;
+}
+/* }}} */
 
 /* {{{ php_load_extension
  */
@@ -93,8 +110,9 @@ PHPAPI int php_load_extension(char *filename, int type, int start_now)
 	char *libpath;
 	zend_module_entry *module_entry;
 	zend_module_entry *(*get_module)(void);
-	int error_type;
+	int error_type, slash_suffix = 0;
 	char *extension_dir;
+	char *err1, *err2;
 
 	if (type == MODULE_PERSISTENT) {
 		extension_dir = INI_STR("extension_dir");
@@ -117,9 +135,9 @@ PHPAPI int php_load_extension(char *filename, int type, int start_now)
 		}
 		libpath = estrdup(filename);
 	} else if (extension_dir && extension_dir[0]) {
-		int extension_dir_len = (int)strlen(extension_dir);
-
-		if (IS_SLASH(extension_dir[extension_dir_len-1])) {
+		slash_suffix = IS_SLASH(extension_dir[strlen(extension_dir)-1]);
+		/* Try as filename first */
+		if (slash_suffix) {
 			spprintf(&libpath, 0, "%s%s", extension_dir, filename); /* SAFE */
 		} else {
 			spprintf(&libpath, 0, "%s%c%s", extension_dir, DEFAULT_SLASH, filename); /* SAFE */
@@ -128,24 +146,31 @@ PHPAPI int php_load_extension(char *filename, int type, int start_now)
 		return FAILURE; /* Not full path given or extension_dir is not set */
 	}
 
-	/* load dynamic symbol */
-	handle = DL_LOAD(libpath);
+	handle = php_load_shlib(libpath, &err1);
 	if (!handle) {
-#ifdef PHP_WIN32
-		char *err = GET_DL_ERROR();
-		if (err && (*err != '\0')) {
-			php_error_docref(NULL, error_type, "Unable to load dynamic library '%s' - %s", libpath, err);
-			LocalFree(err);
+		/* Now, consider 'filename' as extension name and build file name */
+		char *orig_libpath = libpath;
+
+		if (slash_suffix) {
+			spprintf(&libpath, 0, "%s" PHP_SHLIB_EXT_PREFIX "%s." PHP_SHLIB_SUFFIX, extension_dir, filename); /* SAFE */
 		} else {
-			php_error_docref(NULL, error_type, "Unable to load dynamic library '%s' - %s", libpath, "Unknown reason");
+			spprintf(&libpath, 0, "%s%c" PHP_SHLIB_EXT_PREFIX "%s." PHP_SHLIB_SUFFIX, extension_dir, DEFAULT_SLASH, filename); /* SAFE */
 		}
-#else
-		php_error_docref(NULL, error_type, "Unable to load dynamic library '%s' - %s", libpath, GET_DL_ERROR());
-		GET_DL_ERROR(); /* free the buffer storing the error */
-#endif
-		efree(libpath);
-		return FAILURE;
+
+		handle = php_load_shlib(libpath, &err2);
+		if (!handle) {
+			php_error_docref(NULL, error_type, "Unable to load dynamic library '%s' (tried: %s (%s), %s (%s))",
+				filename, orig_libpath, err1, libpath, err2);
+			efree(orig_libpath);
+			efree(err1);
+			efree(libpath);
+			efree(err2);
+			return FAILURE;
+		}
+		efree(orig_libpath);
+		efree(err1);
 	}
+
 	efree(libpath);
 
 	get_module = (zend_module_entry *(*)(void)) DL_FETCH_SYMBOL(handle, "get_module");
